INSTANCE_ATTRIBUTE = 'INSTANCE_ATTRIBUTE'
''' Indicates an option that corresponds to a command object's instance attribute '''

METHOD_NAMED_ARG = 'METHOD_NAMED_ARG'
''' Indicates an option that corresponds to a method's named parameter '''

METHOD_NARGS = 'METHOD_NARGS'
''' Indicates an option that corresponds to the variadic argument consumer of a method (e.g. ``*args``) '''

METHOD_KWARGS = 'METHOD_KWARGS'
''' Indicates an option that corresponds to the keyword argument consumer of a method (e.g. ``**kwargs``) '''
